/* Thread-specific data key creation.
   Copyright (C) 1999 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <errno.h>
#include <limits.h>
#include <pthread.h>
#include <stdlib.h>
#include <spin-lock.h>

#include <pt-internal.h>

/* Create a key value identifying a location in the thread-specific
   data area.  On success the new key is stored in *KEY.  DESTRUCTOR,
   if non-NULL, is called with the value associated to that key at
   thread exit, if the associated value is non-NULL.  */
int pthread_key_create (pthread_key_t * key, void (*destructor) (void *))
{
	int i;
	int num_keys;

	__pthread_mutex_lock (&__pthread_key_lock);

	for (i = 0; i < __pthread_num_keys; i++) {
		if (__pthread_key_destructors[i] == PTHREAD_KEY_AVAILABLE) {
			*key = i;
			__pthread_key_destructors[i] = destructor;
			__pthread_mutex_unlock (&__pthread_key_lock);
			return 0;
		}
	}

	if (__pthread_num_keys == 0)
		num_keys = _POSIX_THREAD_KEYS_MAX;
	else
		num_keys = 2 * __pthread_num_keys;

	__pthread_key_destructors =
		realloc (__pthread_key_destructors,
						 num_keys * sizeof (pthread_key_destructor_t));
	if (__pthread_key_destructors == NULL) {
		__pthread_mutex_unlock (&__pthread_key_lock);
		return ENOMEM;
	}

	for (i = __pthread_num_keys; i < num_keys; i++)
		__pthread_key_destructors[i] = PTHREAD_KEY_AVAILABLE;
	*key = __pthread_num_keys;
	__pthread_key_destructors[__pthread_num_keys] = destructor;
	__pthread_num_keys = num_keys;

	__pthread_mutex_unlock (&__pthread_key_lock);
	return 0;
}
