/* Thread creation.
   Copyright (C) 1999 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <spin-lock.h>

#include <pt-attr.h>
#include <pt-internal.h>

/* The total number of pthreads currently active.  */
int __pthread_total = 1;

/* A lock for pthread_total.  */
__spin_lock_t __pthread_total_lock = __SPIN_LOCK_INITIALIZER;


/* The entry-point for new threads.  */
static void entry_point (void *(*start_routine) (void *), void *arg)
{
	pthread_exit (start_routine (arg));
}

/* Create a thread with attributes given by ATTR, executing
   START_ROUTINE with argument ARG.  */
int
pthread_create (pthread_t * thread, const pthread_attr_t * attr,
		void *(*start_routine) (void *), void *arg)
{
	struct __pthread *setup;
	int err;

	/* Allocate new thread structure.  */
	err = __pthread_alloc (&setup);
	if (err)
		return err;

	/* Copy attributes. */
	memcpy (&setup->attr, attr ? *attr : &__pt_default_attr,
			sizeof (struct __pthread_attr));

	/* Initialize thread state.  */
	setup->state |= (setup->attr.detachstate == PTHREAD_CREATE_DETACHED
			? PTHREAD_DETACHED : 0);

	/* Allocate stack if none was provided.  */
/* XXX
	if (setup->attr.stackaddr == __pt_default_attr.stackaddr) {
		err = __pt_stack_alloc (&setup->attr.stackaddr,
				setup->attr.stacksize);
		if (err)
			goto fail;

		setup->flags |= PTHREAD_STACK;
	}
*/
	/* Allocate stack if none was provided.  */
	/* Start the new thread.  */
	err = __pthread_start (setup, &entry_point, start_routine, arg);
	if (err)
		goto fail;

	/* No critical section since it is impossible that this thread will
	   try to lock the newly created thread again before we return.  */
	__spin_lock (&setup->lock);
	__spin_lock (&__pthread_total_lock);
	__pthread_total++;
	setup->flags |= PTHREAD_COUNTED;
	__spin_unlock (&__pthread_total_lock);
	__spin_unlock (&setup->lock);

	*thread = (pthread_t) setup;
	return 0;

fail:
	if (setup->flags & PTHREAD_STACK)
		__pt_stack_dealloc (setup->attr.stackaddr, setup->attr.stacksize);
	__pthread_dealloc (setup);

	return err;
}
