/* Thread termination.
   Copyright (C) 1999 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <error.h>
#include <ihash.h>
#include <pthread.h>
#include <stdlib.h>
#include <spin-lock.h>

#include <pt-internal.h>

/* XXX */
#define __ihash_find	ihash_find
#define __ihash_free	ihash_free
#define __ihash_iterate	ihash_iterate
#define __ihash_remove  ihash_remove

/* Call the cancellation cleanup handlers on the list CLEANUP.  */
static void internal_function do_cleanup (struct __pthread *thread)
{
	int oldstate;

	pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, &oldstate);

	while (thread->cleanup)
		_pthread_cleanup_pop (thread->cleanup, 1);

	pthread_setcancelstate (oldstate, NULL);
}


/* Table of key destructor functions.  */
pthread_key_destructor_t *__pthread_key_destructors = NULL;
int __pthread_num_keys = 0;

/* XXX */
#define PTHREAD_RECURSIVE_MUTEX_INITIALIZER PTHREAD_MUTEX_INITIALIZER

/* Note that the destructors are allowed to call pthread_key_delete!.
   That's why this is a recursive mutex.  */
pthread_mutex_t __pthread_key_lock = PTHREAD_RECURSIVE_MUTEX_INITIALIZER;

/* List of thread-specific data hashes that are in use.  */
struct pthread_specific *__pthread_specific = NULL;
__spin_lock_t __pthread_specific_lock = __SPIN_LOCK_INITIALIZER;

/* Call the thread-specific data destructor functions for SPECIFIC.  */
static void internal_function
destroy_specific (struct pthread_specific *specific)
{
	void *value;
	int i;

	error_t check (void *value) {
		/* Just a random error value, that feels appropriate.  */
		return EAGAIN;
	}
	/* Remove hash from the list of thread-specific data hashes that are
	   in use.  */ __spin_lock (&__pthread_specific_lock);

	if (specific->next)
		specific->next->prevp = specific->prevp;
	*specific->prevp = specific->next;
	__spin_unlock (&__pthread_specific_lock);

	__pthread_mutex_lock (&__pthread_key_lock);
	while (__ihash_iterate (specific->ht, check) == EAGAIN) {
		for (i = 0; i < __pthread_num_keys; i++) {
			if (__pthread_key_destructors[i] != PTHREAD_KEY_AVAILABLE &&
					__pthread_key_destructors[i] != NULL &&
					(value = __ihash_find (specific->ht, i))) {
				__ihash_remove (specific->ht, i);
				__pthread_key_destructors[i] (value);
			}
		}
	}
	__pthread_mutex_unlock (&__pthread_key_lock);

	__ihash_free (specific->ht);
	free (specific);
}


/* Terminate the current thread and make STATUS available to any
   thread that might join us.  */
void pthread_exit (void *status)
{
	struct __pthread *self = (struct __pthread *) __pthread_self ();

	do_cleanup (self);

	if (self->specific)
		destroy_specific (self->specific);


	__spin_lock (&self->lock);
	__spin_lock (&__pthread_total_lock);
	if (__pthread_total == 1 && self->flags & PTHREAD_COUNTED) {
		/* We are the last thread.  */
		__spin_unlock (&__pthread_total_lock);
		__spin_unlock (&self->lock);
		exit (0);
	} else
		__pthread_total--;
	__spin_unlock (&__pthread_total_lock);
	__spin_unlock (&self->lock);

	__pthread_mutex_lock (&self->state_lock);
	if (self->state != PTHREAD_DETACHED) {
		self->state = PTHREAD_EXITED;
		self->status = status;
	}

	/* Broadcast the condition under all circumstances.  Hopefully, this
	   will make threads that try to join us when we're already detached
	   die horribly instead of blocking indefinitely.  */
	pthread_cond_broadcast (&self->exited);

	__pthread_terminate (self);
}
