/* Internal defenitions for pthreads library.
   Copyright (C) 1999 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#ifndef _PT_INTERNAL_H
#define _PT_INTERNAL_H	1

#include <stddef.h>
#include <spin-lock.h>

#include <pt-sysdep.h>


/* The total number of pthreads currently active.  */
extern int __pthread_total;

/* A lock for pthread_total.  */
extern __spin_lock_t __pthread_total_lock;

extern void __pthread_enqueue (struct __pthread **__queue, struct __pthread *__thread);
extern struct __pthread *__pthread_dequeue (struct __pthread *__queue);
extern void __pthread_remove (struct __pthread *__thread);

/* Initialize the pthreads library.  Returns a new stack pointer
   for the current thread, or NULL of no new stack is needed.  */
extern void *__pt_initialize (void);

/* Alocate a new thread structure.  */
extern int __pthread_alloc (struct __pthread **__thread);

/* Deallocate THREAD.  */
extern void __pthread_dealloc (struct __pthread *__thread);

/* Initialize THREAD.  */
extern int __pthread_init (struct __pthread *__thread);

/* Destroy THREAD.  */
extern void __pthread_destroy (struct __pthread *__thread);

/* Start THREAD, calling ENTRY_POINT with the arguments START_ROUTINE
   and ARG.  */
extern int __pthread_start (struct __pthread *__thread,
			    void (*__entry_point)(void *(*)(void *), void *),
			    void *(*__start_routine)(void *), void *__arg);

/* Terminate THREAD, releasing as much resources as possible at this
   stage.  This function should be called with THREAD->state_lock
   locked.  It will be released before the thread terminates.  */
extern void __pthread_terminate (struct __pthread *__thread);

/* Block THREAD.  */
extern void __pt_block (struct __pthread *__thread);

/* Block THREAD.  */
extern int __pt_timedblock (struct __pthread *__thread,
			    const struct timespec *__abstime);

/* Wakeup THREAD.  */
extern void __pt_wakeup (struct __pthread *__thread);

/* Allocate a new stack of size STACKSIZE.  If successfull, store the
   address of the newly allocated stack in *STACKADDR and return 0.
   Otherwise return an error code (EINVAL for an invalid stack size,
   EAGAIN if the system lacked the necessary resources to allocate a
   new stack).  */
extern int __pt_stack_alloc (void **__stackaddr, size_t __stacksize);

/* Deallocate the stack secribed by STACKADDR and STACKSIZE.  */
extern void __pt_stack_dealloc (void *__stackaddr, size_t __stacksize);

/* Set up the stack for THREAD, such that it appears as if
   START_ROUTINE and ARG were passed to the new thread's entry-point.
   Return the stack pointer for the new thread.  */
extern void *__pt_stack_setup (struct __pthread *__thread,
			       void *(*__start_routine)(void *), void *__arg);

/* Thread-specific data.  */

/* Handy type for key destructor functions.  */
typedef void (*pthread_key_destructor_t) (void *);

/* Table of key destructor functions.  */
extern pthread_key_destructor_t *__pthread_key_destructors;

/* The number of entries in this table.  */
extern int __pthread_num_keys;

/* And a lock for that table.  */
extern pthread_mutex_t __pthread_key_lock;

/* Special value that indicates that a key is available.  */
#define PTHREAD_KEY_AVAILABLE ((pthread_key_destructor_t) -1)


/* This structure describes an element in the list of thread-specefic
   data hashes.  */
struct pthread_specific
  {
    struct ihash *ht;
    struct pthread_specific *next, **prevp;
  };

/* List of thread-specific data hashes that are in use.  */
extern struct pthread_specific *__pthread_specific;

/* A lock for __pthread_specific.  */
extern __spin_lock_t __pthread_specific_lock;


/* And finally the thread itself.  */

#ifndef PTHREAD_SYSDEP_MEMBERS
# define PTHREAD_SYSDEP_MEMBERS
#endif

/* Thread state.  */
enum pthread_state
{
  PTHREAD_DETACHED = 1,
  PTHREAD_EXITED
};

/* Thread flags.  */
enum pthread_flags
{
  PTHREAD_STACK = 1,		/* The stack is our responsibility.  */
  PTHREAD_COUNTED = 2,		/* We contribute to the grand total.  */
};

/* This structure describes a POSIX thread.  */
struct __pthread
  {
    __spin_lock_t lock;
    struct __pthread_attr attr;
    enum pthread_flags flags;
    void *status;
    enum pthread_state state;
    pthread_mutex_t state_lock;
    pthread_cond_t exited;
    struct pthread_specific *specific;
    struct __pthread_cleanup *cleanup;
    enum __pthread_cancelstate cancelstate;
    enum __pthread_canceltype canceltype;
    int cancel;
    PTHREAD_SYSDEP_MEMBERS
    struct __pthread *next, **prevp;
  };

/* The first (main) thread.  */
extern struct __pthread __pthread_main;

#endif /* pt-internal.h */
