/* Wait on a condition.
   Copyright (C) 1999 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <pthread.h>
#include <spin-lock.h>

#include <pt-internal.h>

/* Block on condition variable COND.  MUTEX should be held by the
   calling thread.  On success, MUTEX will be held by the calling
   thread.  */
int pthread_cond_wait (pthread_cond_t *cond, pthread_mutex_t *mutex)
{
	struct __pthread *self = (struct __pthread *) __pthread_self ();
	int oldtype;

	void cleanup (void *arg) {
		int resignal = 0;

		/* Remove ourselves from the list of waiters if we're still on
		   it.  Otherwise, consume the condition signal and signal the
		   condition again.  */
		  __spin_lock (&cond->__lock);
		if (self->prevp) {
			__pthread_remove (self);
			__spin_unlock (&cond->__lock);
		} else {
			__spin_unlock (&cond->__lock);
			resignal = !(__pt_timedblock (self, NULL) == ETIMEDOUT);
		}

		__pthread_mutex_lock (mutex);

		if (resignal)
			pthread_cond_signal (cond);
	}

	pthread_cleanup_push (cleanup, NULL);

	/* Add ourselves to the list of waiters.  */
	__spin_lock (&cond->__lock);
	__pthread_enqueue (&cond->__queue, self);
	__spin_unlock (&cond->__lock);

	__pthread_mutex_unlock (mutex);

	/* This function is a cancellation point.  */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);
	__pt_block (self);
	pthread_setcanceltype (oldtype, NULL);

	pthread_cleanup_pop (0);

	__pthread_mutex_lock (mutex);

	return 0;
}

/* Block on condition variable COND.  MUTEX should be held by the
   calling thread.  On success, MUTEX will be held by the calling
   thread.  */
int
pthread_cond_timedwait (pthread_cond_t *cond, pthread_mutex_t *mutex,
		const struct timespec *abstime)
{
	struct __pthread *self = (struct __pthread *) __pthread_self ();
	int oldtype;
	int err;

	void cleanup (void *arg) {
		int resignal = 0;

		/* Remove ourselves from the list of waiters if we're still on
		   it.  Otherwise, consume the condition signal and signal the
		   condition again.  */
		  __spin_lock (&cond->__lock);
		if (self->prevp) {
			__pthread_remove (self);
			__spin_unlock (&cond->__lock);
		} else {
			__spin_unlock (&cond->__lock);
			resignal = !(__pt_timedblock (self, NULL) == ETIMEDOUT);
		}

		__pthread_mutex_lock (mutex);

		if (resignal)
			pthread_cond_signal (cond);
	}

	pthread_cleanup_push (cleanup, NULL);

	/* Add ourselves to the list of waiters.  */
	__spin_lock (&cond->__lock);
	__pthread_enqueue (&cond->__queue, self);
	__spin_unlock (&cond->__lock);

	__pthread_mutex_unlock (mutex);

	/* This function is a cancellation point.  */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);
	err = __pt_timedblock (self, abstime);
	pthread_setcanceltype (oldtype, NULL);

	pthread_cleanup_pop (0);

	if (err == ETIMEDOUT) {
		/* Remove ourselves from the list of waiters if we're still on
		   it.  Otherwise, consume the condition signal and pretend we
		   didn't time out.  */
		__spin_lock (&cond->__lock);
		if (self->prevp) {
			__pthread_remove (self);
			__spin_unlock (&cond->__lock);
		} else {
			__spin_unlock (&cond->__lock);
			err = __pt_timedblock (self, NULL);
		}
	}

	__pthread_mutex_lock (mutex);

	return err;
}
