/* Allow programs to add items to the atfork hooks. Hurd version.
   Copyright (C) 1999 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <errno.h>
#include <pthread.h>

#include <pt-internal.h>

struct handler_list {
	void (*handler)(void);
	struct handler_list *next, *perv;
};

static pthread_mutex_t __pthread_atfork_lock = PTHREAD_MUTEX_INITIALIZER;
static struct handler_list *atfork_prepare_tail = NULL;
static struct handler_list *atfork_prepare_head = NULL;
static struct handler_list *atfork_parent_tail = NULL;
static struct handler_list *atfork_parent_head = NULL;
static struct handler_list *atfork_child_tail = NULL;
static struct handler_list *atfork_child_head = NULL;

/* register these functions to be called during a fork() */
static void call_atfork_prepare_hooks (void) {
	register struct handler_list *list;
	for (list = atfork_prepare_head; list; list = list->next)
		(list->handler) ();
}
text_set_element (_hurd_fork_prepare_hook, call_atfork_prepare_hooks);

static void call_atfork_parent_hooks (void) {
	register struct handler_list *list;
	for (list = atfork_parent_head; list; list = list->next)
		(list->handler) ();
}
text_set_element (_hurd_fork_parent_hook, call_atfork_parent_hooks);

static void call_atfork_child_hooks (void) {
	register struct handler_list *list;
	for (list = atfork_child_head; list; list = list->next)
		(list->handler) ();
}
text_set_element (_hurd_fork_child_hook, call_atfork_child_hooks);

/* Install handlers to be called when a new process is created with FORK.
   The PREPARE handler is called in the parent process just before performing
   FORK. The PARENT handler is called in the parent process just after FORK.
   The CHILD handler is called in the child process.  Each of the three
   handlers can be NULL, meaning that no handler needs to be called at that
   point.
   PTHREAD_ATFORK can be called several times, in which case the PREPARE
   handlers are called in LIFO order (last added with PTHREAD_ATFORK,
   first called before FORK), and the PARENT and CHILD handlers are called
   in FIFO (first added, first called).  */

int pthread_atfork (void (*prepare)(void), void (*parent)(void),
		void (*child)(void))
{
	struct hander_list *handler;

	handler = (struct handler_list *)
			malloc (sizeof (struct handler_list) * 3);
	if (!handler)
		return ENOMEM;

	handler[0].handler = prepare;
	handler[1].handler = parent;
	handler[2].handler = child;

	/* add handlers to the respective lists so that ...*/

	pthread_mutex_lock(&__pthread_atfork_lock);

	/* prepare handlers are called in LIFO order */
	if (atfork_prepare_tail == NULL) {
		atfork_prepare_tail = atfork_prepare_head = &handler[0];
		handler[0].next = handler[0].prev = NULL;
	} else {
		handler[0].prev = NULL;
		handler[0].next = atfork_prepare_head;
		atfork_prepare_head = &handler[0];
	}
	/* parent handlers are called in FIFO order */
	if (atfork_parent_head == NULL) {
		atfork_parent_tail = atfork_parent_head = &handler[0];
		handler[1].next = handler[1].prev = NULL;
	} else {
		handler[1].next = NULL;
		handler[1].prev = atfork_parent_tail;
		atfork_parent_tail = &handler[1];
	}
	/* child handlers are called in FIFO order */
	if (atfork_child_tail == NULL) {
		atfork_child_tail = atfork_child_head = &handler[0];
		handler[2].next = handler[2].prev = NULL;
	} else {
		handler[2].next = NULL;
		handler[2].prev = atfork_child_tail;
		atfork_child_tail = &handler[2];
	}

	pthread_mutex_unlock(&__pthread_atfork_lock);

	return 0;
}
