/* Initialize a thread structure.  Hurd version.
   Copyright (C) 1999 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <hurd.h>
#include <hurd/signal.h>
#include <hurd/sigpreempt.h>

#include <pt-internal.h>


/* The following code lies at the hart of the Hurd-specific thread
   cancelation code.  Since thread cancelation is so similar to signal
   handling (cancelation requests are acted upon asynchronously and
   acting upon cancelation requests has the same side-effects as
   running a signal handler) we actually use the Hurd's signal
   delivery code to deliver a cancelation request.  For this purpose
   we use SIGKILL, with a sigcode pointing at the thread structure.
   This should be fine since:

   1. The SIGKILL signal cannot be blocked.

   2. The SIGKILL signal will never be sent to another program by any
      program that uses the normal library functions to send a signal.

   Programs that directly use the msg_sig_post RPC may loose in the
   unlikely case that they happen to guess the location of the thread
   structure and manage to direct it at the right thread.  They will
   cancel the thread, but why are they posting a SIGKILL anyway?  */

/* Act upon a cancelation request.  */
static void handler (int signo)
{
	pthread_exit (PTHREAD_CANCELED);
	abort ();
}

/* Decide if the cancelation request has to be acted upon.  */
static sighandler_t
preemptor (struct hurd_signal_preemptor *preemptor, struct hurd_sigstate *ss,
		int *signo, struct hurd_signal_detail *detail)
{
	struct __pthread *thread = (struct __pthread *) detail->code;

	/* Check if the request was indeed directed to us.  */
	if (thread->kernel_thread != ss->thread)
		return SIG_ERR;

	if (thread->canceltype == PTHREAD_CANCEL_ASYNCHRONOUS
			&& thread->cancelstate == PTHREAD_CANCEL_ENABLE)
		return preemptor->handler;

	/* Asynchronous cancelation was disabled.  Set the thread's
	   cancelation flag.  When asynchronous cancelation is enabled or
	   when the target thread reaches a cancelation point it will notice
	   that this flag is set and act upon the cancelation.  */
	thread->cancel = 1;

	return SIG_IGN;
}


/* Initialize THREAD.  */
int __pthread_init (struct __pthread *thread)
{
	struct hurd_sigstate *ss;
	error_t err;

	/* Build wakeup message.  */
	thread->wakeupmsg.msgh_bits = MACH_MSGH_BITS (MACH_MSG_TYPE_COPY_SEND, 0);
	thread->wakeupmsg.msgh_size = 0;
	thread->wakeupmsg.msgh_remote_port = __mach_reply_port ();
	thread->wakeupmsg.msgh_local_port = MACH_PORT_NULL;
	thread->wakeupmsg.msgh_seqno = 0;
	thread->wakeupmsg.msgh_id = 0;

	err = __mach_port_insert_right (__mach_task_self (),
			thread->wakeupmsg.msgh_remote_port,
			thread->wakeupmsg.msgh_remote_port,
			MACH_MSG_TYPE_MAKE_SEND);
	if (err)
		return EAGAIN;

	if (thread != &__pthread_main) {
		/* Create kernel thread.  */
		err = __thread_create (__mach_task_self (),
				&thread->kernel_thread);
		if (err)
			return EAGAIN;

		/* Build preemptor for cancelation.  */
		thread->preemptor.signals = sigmask (SIGKILL);
		thread->preemptor.first = (unsigned long int) thread;
		thread->preemptor.last = (unsigned long int) thread;
		thread->preemptor.preemptor = preemptor;
		thread->preemptor.handler = handler;

		/* And initialize the signal state.  */
		ss = _hurd_thread_sigstate (thread->kernel_thread);
		__spin_lock (&ss->lock);
		sigprocmask (0, NULL, &ss->blocked);
		assert (ss->preemptors == NULL);
		ss->preemptors = &thread->preemptor;
		__spin_unlock (&ss->lock);

		/* XXX If a signal arrives for THREAD, before it starts running,
		   some very bad things may happen!  Right now this cannot happen,
		   since all signals directed at the process will be handled by the
		   first thread (which is already running), but once that's fixed
		   something will have to be done about it.  */
	}

	return 0;
}
